'use strict';

module.exports = {
  jscs: {
    dirnames: ['./lib/async.js', './perf/**', './test/**/*.js', './gulp/**']
  },
  jsbeautifier: {
    dirnames: ['./lib/async.js', './perf/**', './test/**/*.js', './gulp/**']
  }
};
